﻿using System;
using UnityEngine;
using System.Collections.Generic;
using System.Linq;

namespace ky1001
{
    public class L_R_Swap : MVRScript
    {
        #region PluginInfo    
        public string pluginAuthor = "ky1001";
        public string pluginName = "L-R_Swap";
        public string pluginVersion = "4.0";
        public string pluginDate = "2022-08-13";
        public string pluginDescription = "";
        #endregion

        #region Vars
        protected Atom CurrentAtom;
        private Vector3 tempPosition;
        private Vector3 temprPosition;
        private Vector3 templPosition;
        private Quaternion tempRotation;
        private Quaternion temprRotation;
        private Quaternion templRotation;
        private Vector3 tempPosition_h;
        public string tempPosi_r;
        public string tempRota_r;
        public string tempPosi_l;
        public string tempRota_l;
        protected UIDynamicButton button;
        protected UIDynamicButton ubutton;
        protected UIDynamicButton ibutton;
        protected JSONStorableBool toggle;
        protected JSONStorableBool toggle0;
        Dictionary<string, string> PosiDic = new Dictionary<string, string>();
        Dictionary<string, string> RotaDic = new Dictionary<string, string>();
        Dictionary<string, string> AtomDic = new Dictionary<string, string>();
        Dictionary<string, string> LinkDic = new Dictionary<string, string>();
        Dictionary<string, FreeControllerV3.PositionState> PosiST = new Dictionary<string, FreeControllerV3.PositionState>();
        Dictionary<string, FreeControllerV3.RotationState> RotaST = new Dictionary<string, FreeControllerV3.RotationState>();
        Dictionary<string, Rigidbody> RigidbodyList = new Dictionary<string, Rigidbody>();
        Dictionary<string, float> HandPose = new Dictionary<string, float>();
        Dictionary<string, Vector3> PosiV3 = new Dictionary<string, Vector3>();
        Dictionary<string, Quaternion> RotaV3 = new Dictionary<string, Quaternion>();
        Dictionary<string, Vector3> rigidPosiDic = new Dictionary<string, Vector3>();
        Dictionary<string, Quaternion> rigidRotaDic = new Dictionary<string, Quaternion>();
        public string RigidbodyName;
        Rigidbody linkToAtom;

        protected JSONStorableStringChooser atomJSON;
        protected UIDynamicPopup dp;
        private UIDynamicButton rfl_button;
        private bool canUndo = false;
        #endregion

        public readonly string[] ControlNames = new string[]
        {
            "chestControl",
            "headControl",
            "neckControl",
            "pelvisControl",
            "hipControl",
            "abdomenControl",
            "abdomen2Control",
            "eyeTargetControl",
            "penisBaseControl",
            "penisMidControl",
            "penisTipControl",
        };

        public readonly string[] lrControlNames = new string[]
        {
            "HandControl",
            "FootControl",
            "NippleControl",
            "ElbowControl",
            "KneeControl",
            "ToeControl",
            "ThighControl",
            "ArmControl",
            "ShoulderControl",
        };

        private readonly string[] OnControlNames = new string[]
        {
            "chestControl",
            "headControl",
            "hipControl",
            "rHandControl",
            "lHandControl",
            "rFootControl",
            "lFootControl"
        };

        private readonly string[] lrPoseNames = new string[]
        {
            "Hand Chop",
            "Hand Fist",
            "Hand Grasp",
            "Hand Spread",
            "Hand Straighten",
            "Fingers Fist",
            "Fingers Grasp",
            "Fingers In-Out",
            "Fingers Squeeze",
            "Fingers Straighten",
            "Index Finger Bend",
            "Mid Finger Bend",
            "Pinky Finger Bend",
            "Ring Finger Bend",
            "Thumb Bend",
            "Thumb Fist",
            "Thumb Grasp",
            "Thumb In-Out",
            "Thumb Straighten",
            "Big Toe Bend",
            "Index Toe Bend",
            "Mid Toe Bend",
            "Pinky Toe Bend",
            "Ring Toe Bend",
            "Toes Fan Down",
            "Toes Small Curl",
            "Toes Spread",
        };

        private readonly string[] rigidbodyNames = new string[]
        {
            "hip",
            "pelvis",

            "rThigh",
            "rShin",
            "rFoot",
            "rToe",

            "lThigh",
            "lShin",
            "lFoot",
            "lToe",

            "abdomen",
            "abdomen2",

            "chest",

            "rCollar",
            "rShldr",
            "rForeArm",
            "rHand",

            "lCollar",
            "lShldr",
            "lForeArm",
            "lHand",

            "neck",
            "head"
        };

        // Use this for initialization
        public override void Init()
        {
            try
            {
                #region Buttons
                atomJSON = new JSONStorableStringChooser("Target", null, "None", "Target", clearUndo);
                RegisterStringChooser(atomJSON);
                dp = CreateScrollablePopup(atomJSON, false);
                dp.popupPanelHeight = 650f;
                dp.popup.selectColor = new Color(0.8f, 0.8f, 0.8f);

                mkList();

                var spacer = CreateSpacer(true);
                spacer.height = 30f;

                rfl_button = CreateButton("Refresh List", true);
                if (rfl_button != null)
                {
                    rfl_button.button.onClick.AddListener(mkList);
                }
                JSONStorableAction refreshListAct = new JSONStorableAction("Refresh List", () => { mkList(); });
                RegisterAction(refreshListAct);

                spacer = CreateSpacer(true);
                spacer.height = 35f;

                spacer = CreateSpacer(false);
                spacer.height = 30f;

                toggle0 = new JSONStorableBool("World Coordinate", false);
                CreateToggle(toggle0, true);
                RegisterBool(toggle0);

                toggle = new JSONStorableBool("Keep Link To Other Atom", true);
                CreateToggle(toggle, true);
                RegisterBool(toggle);

                button = CreateButton("L-R Swap");
                if (button != null)
                {
                    button.button.onClick.AddListener(lrRevButtonCallback);
                }
                JSONStorableAction lrRevAct = new JSONStorableAction("L-R Swap", () => { lrRevButtonCallback(); });
                RegisterAction(lrRevAct);

                ubutton = CreateButton("Undo");
                if (button != null)
                {
                    ubutton.button.onClick.AddListener(Undo);
                }
                JSONStorableAction undoAct = new JSONStorableAction("Undo", () => { Undo(); });
                RegisterAction(undoAct);

                ibutton = CreateButton("Status Initialize", true);
                if (ibutton != null)
                {
                    ibutton.button.onClick.AddListener(StatusInit);
                }
                JSONStorableAction statusInitAct = new JSONStorableAction("Status Initialize", () => { StatusInit(); });
                RegisterAction(statusInitAct);
                #endregion

                PosiST["On"] = FreeControllerV3.PositionState.On;
                PosiST["Comply"] = FreeControllerV3.PositionState.Comply;
                PosiST["Off"] = FreeControllerV3.PositionState.Off;
                PosiST["ParentLink"] = FreeControllerV3.PositionState.ParentLink;
                PosiST["PhysicsLink"] = FreeControllerV3.PositionState.PhysicsLink;
                PosiST["Hold"] = FreeControllerV3.PositionState.Hold;
                PosiST["Lock"] = FreeControllerV3.PositionState.Lock;

                RotaST["On"] = FreeControllerV3.RotationState.On;
                RotaST["Comply"] = FreeControllerV3.RotationState.Comply;
                RotaST["Off"] = FreeControllerV3.RotationState.Off;
                RotaST["ParentLink"] = FreeControllerV3.RotationState.ParentLink;
                RotaST["PhysicsLink"] = FreeControllerV3.RotationState.PhysicsLink;
                RotaST["Hold"] = FreeControllerV3.RotationState.Hold;
                RotaST["Lock"] = FreeControllerV3.RotationState.Lock;
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        void Start()
        {
            UpdateInitialMorphs();
        }

        private void clearUndo(string dummy)
        {
            canUndo = false;
        }

        public void mkList()
        {
            List<string> targetChoices = new List<string>();
            foreach (string atomUID in SuperController.singleton.GetAtomUIDs())
            {
                var targetAtom = SuperController.singleton.GetAtomByUid(atomUID);
                if (targetAtom.type == "Person")
                {
                    targetChoices.Add(atomUID);
                }
            }
            atomJSON.choices = targetChoices;
            if (containingAtom.type == "Person")
            {
                atomJSON.val = containingAtom.name;
                CurrentAtom = SuperController.singleton.GetAtomByUid(atomJSON.val);
            }
            else
            {
                atomJSON.val = "None";
            }
        }

        private void UpdateInitialMorphs()
        {
            JSONStorable geometry = CurrentAtom.GetStorableByID("geometry");
            DAZCharacterSelector character = geometry as DAZCharacterSelector;
            GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;
        }

        private void ApplyMorphs()
        {
            JSONStorable geometry = CurrentAtom.GetStorableByID("geometry");
            DAZCharacterSelector character = geometry as DAZCharacterSelector;
            GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;

            List<string> lrPoseList = new List<string>(lrPoseNames);
            foreach (string lrPoseNames in lrPoseList)
            {
                if (morphControl.GetMorphByDisplayName("Right " + lrPoseNames) != null && morphControl.GetMorphByDisplayName("Left " + lrPoseNames) != null)
                {
                    float tempPose_r = morphControl.GetMorphByDisplayName("Right " + lrPoseNames).morphValue;
                    HandPose["Right " + lrPoseNames] = tempPose_r;

                    float tempPose_l = morphControl.GetMorphByDisplayName("Left " + lrPoseNames).morphValue;
                    HandPose["Left " + lrPoseNames] = tempPose_l;

                    morphControl.GetMorphByDisplayName("Right " + lrPoseNames).morphValue = tempPose_l;
                    morphControl.GetMorphByDisplayName("Left " + lrPoseNames).morphValue = tempPose_r;
                }
            }
        }

        public void lrRevButtonCallback()
        {
            if (atomJSON.val != "None")
            {
                CurrentAtom = SuperController.singleton.GetAtomByUid(atomJSON.val);
                canUndo = true;
                Reverse();
                Invoke("StatusRecall", 0.5f);
                Invoke("ColButtonCallback", 0.5f);
            }
        }

        public void Reverse()
        {
            CurrentAtom.collisionEnabledJSON.val = false;

            Rigidbody[] objects = CurrentAtom.rigidbodies;
            foreach (var ObjectName in objects)
            {
                RigidbodyList[ObjectName.ToString()] = ObjectName;
            }

            List<string> ControlList = new List<string>(ControlNames);
            foreach (string ControlNames in ControlList)
            {
                FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;
                PosiDic[ControlNames] = CurrentController.currentPositionState.ToString();
                RotaDic[ControlNames] = CurrentController.currentRotationState.ToString();

                if (!CurrentController.linkToRB)
                {
                    AtomDic[ControlNames] = "None";
                }
                else
                {
                    var jc = CurrentController.GetJSON();
                    var linkTo = jc["linkTo"].Value;
                    AtomDic[ControlNames] = linkTo.Split(":"[0])[0];
                    //SuperController.LogMessage(linkTo.Split(":"[0])[0]);
                }

                if (!CurrentController.linkToRB)
                {
                    LinkDic[ControlNames] = "None";
                    RigidbodyList["None"] = CurrentController.linkToRB;
                }
                else
                {
                    LinkDic[ControlNames] = CurrentController.linkToRB.ToString();
                }

                CurrentController.SetLinkToAtom("None");
                CurrentController.SetLinkToRigidbodyObject("None");
                CurrentController.currentPositionState = PosiST["On"];
                CurrentController.currentRotationState = RotaST["On"];
            }

            List<string> lrControlList = new List<string>(lrControlNames);
            foreach (string lrControlNames in lrControlList)
            {
                FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                PosiDic["r" + lrControlNames] = rCurrentController.currentPositionState.ToString();
                RotaDic["r" + lrControlNames] = rCurrentController.currentRotationState.ToString();

                if (!rCurrentController.linkToRB)
                {
                    AtomDic["r" + lrControlNames] = "None";
                }
                else
                {
                    var jc = rCurrentController.GetJSON();
                    var linkTo = jc["linkTo"].Value;
                    AtomDic["r" + lrControlNames] = linkTo.Split(":"[0])[0];
                }

                if (!rCurrentController.linkToRB)
                {
                    LinkDic["r" + lrControlNames] = "None";
                }
                else
                {
                    LinkDic["r" + lrControlNames] = rCurrentController.linkToRB.ToString();
                }

                rCurrentController.SetLinkToAtom("None");
                rCurrentController.SetLinkToRigidbodyObject("None");
                rCurrentController.currentPositionState = PosiST["On"];
                rCurrentController.currentRotationState = RotaST["On"];

                FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;
                PosiDic["l" + lrControlNames] = lCurrentController.currentPositionState.ToString();
                RotaDic["l" + lrControlNames] = lCurrentController.currentRotationState.ToString();

                if (!lCurrentController.linkToRB)
                {
                    AtomDic["l" + lrControlNames] = "None";
                }
                else
                {
                    var jc = lCurrentController.GetJSON();
                    var linkTo = jc["linkTo"].Value;
                    AtomDic["l" + lrControlNames] = linkTo.Split(":"[0])[0];
                }

                if (!lCurrentController.linkToRB)
                {
                    LinkDic["l" + lrControlNames] = "None";
                }
                else
                {
                    LinkDic["l" + lrControlNames] = lCurrentController.linkToRB.ToString();
                }

                lCurrentController.SetLinkToAtom("None");
                lCurrentController.SetLinkToRigidbodyObject("None");
                lCurrentController.currentPositionState = PosiST["On"];
                lCurrentController.currentRotationState = RotaST["On"];
            }

            foreach (var item in rigidbodyNames)
            {
                var _rigidbody = CurrentAtom.rigidbodies.First(rb => rb.name == item);
                rigidPosiDic[item] = _rigidbody.transform.position;
                rigidRotaDic[item] = _rigidbody.transform.rotation;
            }

            if (toggle0.val == false)
            {
                foreach (string ControlNames in ControlList)
                {
                    FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;

                    if (ControlNames == "chestControl")
                    {
                        tempPosition_h.x = CurrentController.transform.localPosition.x;
                    }

                    PosiV3[ControlNames] = CurrentController.transform.localPosition;
                    RotaV3[ControlNames] = CurrentController.transform.localRotation;

                    tempPosition.x = -CurrentController.transform.localPosition.x + tempPosition_h.x * 2;
                    tempPosition.y = CurrentController.transform.localPosition.y;
                    tempPosition.z = CurrentController.transform.localPosition.z;
                    CurrentController.transform.localPosition = tempPosition;
                    tempRotation.x = CurrentController.transform.localRotation.x;
                    tempRotation.y = -CurrentController.transform.localRotation.y;
                    tempRotation.z = -CurrentController.transform.localRotation.z;
                    tempRotation.w = CurrentController.transform.localRotation.w;
                    CurrentController.transform.localRotation = tempRotation;
                }

                foreach (string lrControlNames in lrControlList)
                {
                    FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;

                    PosiV3["r" + lrControlNames] = CurrentController.transform.localPosition;
                    RotaV3["r" + lrControlNames] = CurrentController.transform.localRotation;

                    temprPosition.x = -CurrentController.transform.localPosition.x + tempPosition_h.x * 2;
                    temprPosition.y = CurrentController.transform.localPosition.y;
                    temprPosition.z = CurrentController.transform.localPosition.z;
                    temprRotation.x = CurrentController.transform.localRotation.x;
                    temprRotation.y = -CurrentController.transform.localRotation.y;
                    temprRotation.z = -CurrentController.transform.localRotation.z;
                    temprRotation.w = CurrentController.transform.localRotation.w;

                    FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;

                    PosiV3["l" + lrControlNames] = lCurrentController.transform.localPosition;
                    RotaV3["l" + lrControlNames] = lCurrentController.transform.localRotation;

                    templPosition.x = -lCurrentController.transform.localPosition.x + tempPosition_h.x * 2;
                    templPosition.y = lCurrentController.transform.localPosition.y;
                    templPosition.z = lCurrentController.transform.localPosition.z;
                    templRotation.x = lCurrentController.transform.localRotation.x;
                    templRotation.y = -lCurrentController.transform.localRotation.y;
                    templRotation.z = -lCurrentController.transform.localRotation.z;
                    templRotation.w = lCurrentController.transform.localRotation.w;
                    lCurrentController.transform.localPosition = temprPosition;
                    lCurrentController.transform.localRotation = temprRotation;

                    FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                    rCurrentController.transform.localPosition = templPosition;
                    rCurrentController.transform.localRotation = templRotation;
                }
            }
            else
            {
                foreach (string ControlNames in ControlList)
                {
                    FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;

                    if (ControlNames == "chestControl")
                    {
                        tempPosition_h.x = CurrentController.transform.position.x;
                    }

                    PosiV3[ControlNames] = CurrentController.transform.position;
                    RotaV3[ControlNames] = CurrentController.transform.rotation;

                    tempPosition.x = -CurrentController.transform.position.x + tempPosition_h.x * 2;
                    tempPosition.y = CurrentController.transform.position.y;
                    tempPosition.z = CurrentController.transform.position.z;
                    CurrentController.transform.position = tempPosition;
                    tempRotation.x = CurrentController.transform.rotation.x;
                    tempRotation.y = -CurrentController.transform.rotation.y;
                    tempRotation.z = -CurrentController.transform.rotation.z;
                    tempRotation.w = CurrentController.transform.rotation.w;
                    CurrentController.transform.rotation = tempRotation;
                }

                foreach (string lrControlNames in lrControlList)
                {
                    FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;

                    PosiV3["r" + lrControlNames] = CurrentController.transform.position;
                    RotaV3["r" + lrControlNames] = CurrentController.transform.rotation;

                    temprPosition.x = -CurrentController.transform.position.x + tempPosition_h.x * 2;
                    temprPosition.y = CurrentController.transform.position.y;
                    temprPosition.z = CurrentController.transform.position.z;
                    temprRotation.x = CurrentController.transform.rotation.x;
                    temprRotation.y = -CurrentController.transform.rotation.y;
                    temprRotation.z = -CurrentController.transform.rotation.z;
                    temprRotation.w = CurrentController.transform.rotation.w;

                    FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;

                    PosiV3["l" + lrControlNames] = lCurrentController.transform.position;
                    RotaV3["l" + lrControlNames] = lCurrentController.transform.rotation;

                    templPosition.x = -lCurrentController.transform.position.x + tempPosition_h.x * 2;
                    templPosition.y = lCurrentController.transform.position.y;
                    templPosition.z = lCurrentController.transform.position.z;
                    templRotation.x = lCurrentController.transform.rotation.x;
                    templRotation.y = -lCurrentController.transform.rotation.y;
                    templRotation.z = -lCurrentController.transform.rotation.z;
                    templRotation.w = lCurrentController.transform.rotation.w;
                    lCurrentController.transform.position = temprPosition;
                    lCurrentController.transform.rotation = temprRotation;

                    FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                    rCurrentController.transform.position = templPosition;
                    rCurrentController.transform.rotation = templRotation;
                }
            }

            foreach (var item in rigidbodyNames)
            {
                var _rigidbody = CurrentAtom.rigidbodies.First(rb => rb.name == item);
                _rigidbody.detectCollisions = false;
                var newItem = "";

                if (item.Substring(0, 1) == "r")
                {
                    newItem = "l" + item.Substring(1, item.Length - 1);
                }
                else if (item.Substring(0, 1) == "l")
                {
                    newItem = "r" + item.Substring(1, item.Length - 1);
                }
                else
                {
                    newItem = item;
                }

                //tempPosition_h = CurrentAtom.rigidbodies.First(rb => rb.name == "chest").transform.position;
                tempPosition_h = CurrentAtom.rigidbodies.First(rb => rb.name == "hip").transform.position;

                tempPosition.x = -rigidPosiDic[newItem].x + tempPosition_h.x * 2;
                tempPosition.y = rigidPosiDic[newItem].y;
                tempPosition.z = rigidPosiDic[newItem].z;
                _rigidbody.transform.position = tempPosition;
                //_rigidbody.transform.position = tempPosition_h;
                tempRotation.x = rigidRotaDic[newItem].x;
                tempRotation.y = -rigidRotaDic[newItem].y;
                tempRotation.z = -rigidRotaDic[newItem].z;
                tempRotation.w = rigidRotaDic[newItem].w;
                _rigidbody.transform.rotation = tempRotation;
            }
            ApplyMorphs();
        }

        public void StatusRecall()
        {
            List<string> ControlList = new List<string>(ControlNames);
            foreach (string ControlNames in ControlList)
            {
                FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;
                CurrentController.currentPositionState = PosiST[PosiDic[ControlNames]];
                CurrentController.currentRotationState = RotaST[RotaDic[ControlNames]];

                if (AtomDic[ControlNames] != "None")
                {
                    if (LinkDic[ControlNames].Substring(0, 1) == "l")
                    {
                        if (AtomDic[ControlNames] == CurrentAtom.uid || toggle.val == false)
                        {
                            RigidbodyName = "r" + LinkDic[ControlNames].Substring(1, LinkDic[ControlNames].Length - 1);
                            linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic[ControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                            CurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                        }
                        else
                        {
                            RigidbodyName = "l" + LinkDic[ControlNames].Substring(1, LinkDic[ControlNames].Length - 1);
                            linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic[ControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                            CurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                        }
                    }
                    else if (LinkDic[ControlNames].Substring(0, 1) == "r")
                    {
                        if (AtomDic[ControlNames] == CurrentAtom.uid || toggle.val == false)
                        {
                            RigidbodyName = "l" + LinkDic[ControlNames].Substring(1, LinkDic[ControlNames].Length - 1);
                            linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic[ControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                            CurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                        }
                        else
                        {
                            RigidbodyName = "r" + LinkDic[ControlNames].Substring(1, LinkDic[ControlNames].Length - 1);
                            linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic[ControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                            CurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                        }
                    }
                }
            }

            List<string> lrControlList = new List<string>(lrControlNames);
            foreach (string lrControlNames in lrControlList)
            {
                FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;
                rCurrentController.SetLinkToAtom(AtomDic["l" + lrControlNames]);
                lCurrentController.SetLinkToAtom(AtomDic["r" + lrControlNames]);
                rCurrentController.currentPositionState = PosiST[PosiDic["l" + lrControlNames]];
                rCurrentController.currentRotationState = RotaST[RotaDic["l" + lrControlNames]];
                lCurrentController.currentPositionState = PosiST[PosiDic["r" + lrControlNames]];
                lCurrentController.currentRotationState = RotaST[RotaDic["r" + lrControlNames]];

                if (LinkDic["l" + lrControlNames] != "None")
                {
                    if (LinkDic["l" + lrControlNames].Substring(0, 1) == "l")
                    {
                        if (AtomDic["l" + lrControlNames] == CurrentAtom.uid || toggle.val == false)
                        {
                            RigidbodyName = "r" + LinkDic["l" + lrControlNames].Substring(1, LinkDic["l" + lrControlNames].Length - 1);
                        }
                        else
                        {
                            RigidbodyName = LinkDic["l" + lrControlNames];
                        }
                    }
                    else if (LinkDic["l" + lrControlNames].Substring(0, 1) == "r")
                    {
                        if (AtomDic["l" + lrControlNames] == CurrentAtom.uid || toggle.val == false)
                        {
                            RigidbodyName = "l" + LinkDic["l" + lrControlNames].Substring(1, LinkDic["l" + lrControlNames].Length - 1);
                        }
                        else
                        {
                            RigidbodyName = LinkDic["l" + lrControlNames];
                        }
                    }
                    else
                    {
                        RigidbodyName = LinkDic["l" + lrControlNames];
                    }
                    linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic["l" + lrControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                    rCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                }
                else
                {
                    linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody("None:None");
                    rCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                }

                if (LinkDic["r" + lrControlNames] != "None")
                {
                    if (LinkDic["r" + lrControlNames].Substring(0, 1) == "l")
                    {
                        if (AtomDic["r" + lrControlNames] == CurrentAtom.uid || toggle.val == false)
                        {
                            RigidbodyName = "r" + LinkDic["r" + lrControlNames].Substring(1, LinkDic["r" + lrControlNames].Length - 1);
                        }
                        else
                        {
                            RigidbodyName = LinkDic["r" + lrControlNames];
                        }
                    }
                    else if (LinkDic["r" + lrControlNames].Substring(0, 1) == "r")
                    {
                        if (AtomDic["r" + lrControlNames] == CurrentAtom.uid || toggle.val == false)
                        {
                            RigidbodyName = "l" + LinkDic["r" + lrControlNames].Substring(1, LinkDic["r" + lrControlNames].Length - 1);
                        }
                        else
                        {
                            RigidbodyName = LinkDic["r" + lrControlNames];
                        }
                    }
                    else
                    {
                        RigidbodyName = LinkDic["r" + lrControlNames];
                    }
                    linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic["r" + lrControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                    lCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                }
                else
                {
                    linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody("None:None");
                    lCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                }
            }
        }

        public void ColButtonCallback()
        {
            CurrentAtom.collisionEnabledJSON.val = true;
        }

        public void Undo()
        {
            if (atomJSON.val != "None" && canUndo == true)
            {
                CurrentAtom = SuperController.singleton.GetAtomByUid(atomJSON.val);
                AllControlOn();
                UndoMain();
                Invoke("StatusRecall2", 0.5f);
                Invoke("ColButtonCallback", 0.5f);
            }
        }

        public void AllControlOn()
        {
            List<string> ControlList = new List<string>(ControlNames);
            foreach (string ControlNames in ControlList)
            {
                FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;
                CurrentController.currentPositionState = PosiST["On"];
                CurrentController.currentRotationState = RotaST["On"];
            }

            List<string> lrControlList = new List<string>(lrControlNames);
            foreach (string lrControlNames in lrControlList)
            {
                FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                rCurrentController.currentPositionState = PosiST["On"];
                rCurrentController.currentRotationState = RotaST["On"];
                FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;
                lCurrentController.currentPositionState = PosiST["On"];
                lCurrentController.currentRotationState = RotaST["On"];
            }
        }

        public void StatusRecall2()
        {
            Rigidbody[] objects = CurrentAtom.rigidbodies;
            foreach (var ObjectName in objects)
            {
                RigidbodyList[ObjectName.ToString()] = ObjectName;
            }

            List<string> ControlList = new List<string>(ControlNames);
            foreach (string ControlNames in ControlList)
            {
                FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;
                CurrentController.currentPositionState = PosiST[PosiDic[ControlNames]];
                CurrentController.currentRotationState = RotaST[RotaDic[ControlNames]];
                RigidbodyName = LinkDic[ControlNames];
                linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic[ControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                CurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
            }

            List<string> lrControlList = new List<string>(lrControlNames);
            foreach (string lrControlNames in lrControlList)
            {
                FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                rCurrentController.currentPositionState = PosiST[PosiDic["r" + lrControlNames]];
                rCurrentController.currentRotationState = RotaST[RotaDic["r" + lrControlNames]];
                RigidbodyName = LinkDic["r" + lrControlNames];
                linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic["r" + lrControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                rCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);

                FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;
                lCurrentController.currentPositionState = PosiST[PosiDic["l" + lrControlNames]];
                lCurrentController.currentRotationState = RotaST[RotaDic["l" + lrControlNames]];
                RigidbodyName = LinkDic["l" + lrControlNames];
                linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic["l" + lrControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                lCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
            }
        }

        public void UndoMain()
        {
            CurrentAtom.collisionEnabledJSON.val = false;

            Rigidbody[] objects = CurrentAtom.rigidbodies;
            foreach (var ObjectName in objects)
            {
                RigidbodyList[ObjectName.ToString()] = ObjectName;
            }

            if (toggle0.val == false)
            {
                List<string> ControlList = new List<string>(ControlNames);
                foreach (string ControlNames in ControlList)
                {
                    FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;
                    CurrentController.transform.localPosition = PosiV3[ControlNames];
                    CurrentController.transform.localRotation = RotaV3[ControlNames];
                }

                List<string> lrControlList = new List<string>(lrControlNames);
                foreach (string lrControlNames in lrControlList)
                {
                    FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                    rCurrentController.transform.localPosition = PosiV3["r" + lrControlNames];
                    rCurrentController.transform.localRotation = RotaV3["r" + lrControlNames];

                    FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;
                    lCurrentController.transform.localPosition = PosiV3["l" + lrControlNames];
                    lCurrentController.transform.localRotation = RotaV3["l" + lrControlNames];
                }
            }
            else
            {
                List<string> ControlList = new List<string>(ControlNames);
                foreach (string ControlNames in ControlList)
                {
                    FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;
                    CurrentController.transform.position = PosiV3[ControlNames];
                    CurrentController.transform.rotation = RotaV3[ControlNames];
                }

                List<string> lrControlList = new List<string>(lrControlNames);
                foreach (string lrControlNames in lrControlList)
                {
                    FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                    rCurrentController.transform.position = PosiV3["r" + lrControlNames];
                    rCurrentController.transform.rotation = RotaV3["r" + lrControlNames];

                    FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;
                    lCurrentController.transform.position = PosiV3["l" + lrControlNames];
                    lCurrentController.transform.rotation = RotaV3["l" + lrControlNames];
                }
            }

            foreach (var item in rigidbodyNames)
            {
                var _rigidbody = CurrentAtom.rigidbodies.First(rb => rb.name == item);
                _rigidbody.detectCollisions = false;

                tempPosition.x = rigidPosiDic[item].x;
                tempPosition.y = rigidPosiDic[item].y;
                tempPosition.z = rigidPosiDic[item].z;
                _rigidbody.transform.position = tempPosition;
                tempRotation.x = rigidRotaDic[item].x;
                tempRotation.y = rigidRotaDic[item].y;
                tempRotation.z = rigidRotaDic[item].z;
                tempRotation.w = rigidRotaDic[item].w;
                _rigidbody.transform.rotation = tempRotation;
            }

            JSONStorable geometry = CurrentAtom.GetStorableByID("geometry");
            DAZCharacterSelector character = geometry as DAZCharacterSelector;
            GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;

            List<string> lrPoseList = new List<string>(lrPoseNames);
            foreach (string lrPoseNames in lrPoseList)
            {
                if (morphControl.GetMorphByDisplayName("Right " + lrPoseNames) != null && morphControl.GetMorphByDisplayName("Left " + lrPoseNames) != null)
                {
                    morphControl.GetMorphByDisplayName("Right " + lrPoseNames).morphValue = HandPose["Right " + lrPoseNames];
                    morphControl.GetMorphByDisplayName("Left " + lrPoseNames).morphValue = HandPose["Left " + lrPoseNames];
                }
            }
        }

        public void StatusInit()
        {
            if (atomJSON.val != "None")
            {
                CurrentAtom = SuperController.singleton.GetAtomByUid(atomJSON.val);

                List<string> ControlList = new List<string>(ControlNames);
                foreach (string ControlNames in ControlList)
                {
                    FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;
                    CurrentController.SetLinkToAtom("None");
                    CurrentController.SetLinkToRigidbodyObject("None");
                    CurrentController.currentPositionState = PosiST["Off"];
                    CurrentController.currentRotationState = RotaST["Off"];
                    linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody("None:None");
                    CurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                }

                List<string> lrControlList = new List<string>(lrControlNames);
                foreach (string lrControlNames in lrControlList)
                {
                    FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                    rCurrentController.SetLinkToAtom("None");
                    rCurrentController.SetLinkToRigidbodyObject("None");
                    rCurrentController.currentPositionState = PosiST["Off"];
                    rCurrentController.currentRotationState = RotaST["Off"];
                    linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody("None:None");
                    rCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);

                    FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;
                    lCurrentController.SetLinkToAtom("None");
                    lCurrentController.SetLinkToRigidbodyObject("None");
                    lCurrentController.currentPositionState = PosiST["Off"];
                    lCurrentController.currentRotationState = RotaST["Off"];
                    linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody("None:None");
                    lCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                }

                List<string> OnControlList = new List<string>(OnControlNames);
                foreach (string OnControlNames in OnControlList)
                {
                    FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(OnControlNames) as FreeControllerV3;
                    CurrentController.currentPositionState = PosiST["On"];
                    CurrentController.currentRotationState = RotaST["On"];
                }
            }
        }
    }
}

